/* ADO in Plain C Demo code by anonytmouse.
 * Please see the readme for more details */

/* This file needs to be linked with oleaut32.lib and ole32.lib */

#include <stdio.h>
#include <windows.h>

/* We define INITGUID so that guids are obtained from the header files
 * instead of a library. Note adoid.h should be included before msado15.h */
#define INITGUID
#include <objbase.h>
#include "adoid.h"
#include "msado15.h"
#undef INITGUID

/* Our custom com helper functions code reside in comHelper.h */
#include "comHelper.h"


/* Prototypes */
void process();
long ADOErrorInfo(ADOConnection * conn);
HRESULT GetFieldNames(ADORecordset * pRs, long  * plngFieldCount, LPTSTR ** strNameArray );
void CleanupStringArray(LPTSTR * strArray, long elementCnt);
void Data2HTML();

/* ----------------------------------------------------------------------------------- */
int main(void) {

	InitComHelper(); /* This includes a call to CoInitialize */

	process(); /* Function that shows data on screen */

	Data2HTML(); /* Function that outputs data to a file in HTML format */

	DestroyComHelper(); /* Includes call to CoUninitialize */

	printf("\nPress Enter to continue...");
	getchar();

	return ERROR_SUCCESS;
}


/* ----------------------------------------------------------------------------------- */
/* Macro to print error details, cleanup and quit if the function fails */
#define ADO_TRY(func) if (func != S_OK) { \
		ADOErrorInfo(conn); \
		goto cleanup; }

/* Macro that reports error details and sets lastADOErrorNum to the error number
 * of the first(and usually only) error in the error collection */
#define ADO_REPORT(func) if (func != S_OK) { \
		lastADOErrorNum = ADOErrorInfo(conn);  } \
		else lastADOErrorNum = 0;

void process() {

	/* This function is the main demo function. It will get data and print it
	 * to the screen. To change the data source, change the connection string
	 * and the select statement. */

	ADOConnection * conn = NULL;
	ADORecordset * rs = NULL; 	/* We set things that need to be cleaned up */
	BSTR bStr = NULL;		/* to NULL so we know if they have been initialised */
	LPTSTR * strArrayFieldNames = NULL;

	HRESULT hr;
	VARIANT_BOOL vblEOF;
	long lastADOErrorNum;
	long lngFieldCnt;
	LONG lngState;

	LPTSTR strTemp;
	VARIANT varFieldIndex, varFieldContents;
	long idx, recordNum = 1;

	/* Create our connection object */
	hr = CoCreateInstance(&CLSID_CADOConnection , NULL ,CLSCTX_INPROC_SERVER,
		&IID_IADOConnection ,(LPVOID*) &conn);

	if (hr != S_OK) return; /* Failed to create connection */

	/* Connect to the data source */
	bStr = SAS(L"Provider=Microsoft.Jet.OLEDB.4.0;Data Source=Whales.xls;Extended Properties=\"Excel 8.0;HDR=Yes;IMEX=1;\"" );
	ADO_TRY( conn->lpVtbl->Open(conn, bStr, bstrEmpty, bstrEmpty, 0) );
	SFS(bStr);

	/* Sample insert statement that demonstrates our error reporting function.
	 * This will generate an error as our connection string is read only. */
	bStr = SAS(L"INSERT INTO [Whales$] (ID,Species,Population,Status) VALUES(22,'Humpback',10000,'No Idea')");
	ADO_REPORT( conn->lpVtbl->Execute(conn, bStr, NULL,adExecuteNoRecords, NULL) );
	SFS(bStr);

	/* Execute a select statement to retrieve a recordset */
	bStr = SAS(L"SELECT ID, Species, Population, Status FROM [Whales$] ORDER BY ID");
	ADO_TRY( conn->lpVtbl->Execute(conn, bStr, NULL,0,&rs) ); /* Last argument receives recordset */
	SFS(bStr);

	/* Check that a valid, open recordset was returned */
	if ( rs == NULL  || rs->lpVtbl->get_State(rs,&lngState) != S_OK
			   || lngState == adStateClosed) {
		/* If a query that returns no data is executed a closed recordset is returned */
		goto cleanup;
	}

	/* Get field names and the number of fields */
	ADO_TRY( GetFieldNames(rs,&lngFieldCnt,&strArrayFieldNames) );

	varFieldIndex.vt = VT_I4; /* Integer type variant */

	/* Loop through the records in the recordset */
	while (TRUE) {

		/* Check if we are at the end of the recordset */
		ADO_TRY( rs->lpVtbl->get_EOF(rs,&vblEOF) );
		if (vblEOF) break; /* No more records */

		printf("\n -----Showing record number %d-----",recordNum);

		/* Loop through each field and print its name and contents */
		for (idx=0;idx < lngFieldCnt;idx++) {

			varFieldIndex.lVal = idx; /* The zero based field index */

			/* get_Collect is a little known, undocumented shortcut to rs.fields.item(idx).value
			 * It allows us to skip getting the fields collection and then each field */
			ADO_REPORT( rs->lpVtbl->get_Collect(rs,varFieldIndex,&varFieldContents) );

			/* lastADOErrorNum is set in the ADO_REPORT macro */
			/* Convert the field contents into a TStr so we can print them */

			if (lastADOErrorNum == 0 && varFieldContents.vt == VT_NULL) strTemp = TEXT("#NULL#");
			else if (lastADOErrorNum == 0) strTemp = Variant2TStr(varFieldContents); /* Variant2TStr is in comHelper.h */
			else strTemp = NULL; /* Error */

			printf("\n %s: %s",strArrayFieldNames[idx],(strTemp ? strTemp : TEXT("#ERROR#") ) );

			if ( strTemp != NULL && varFieldContents.vt != VT_NULL) GlobalFree(strTemp);
			VariantClear(&varFieldContents);
		}

		/* Move to the next record */
		ADO_TRY( rs->lpVtbl->MoveNext(rs) );

		recordNum++;
	}


cleanup:

	if (strArrayFieldNames != NULL) CleanupStringArray(strArrayFieldNames,lngFieldCnt);

	SFS(bStr);

	if (conn != NULL) {
		hr = conn->lpVtbl->get_State(conn,&lngState);
		/* Closing the connection also closes active recordsets associated with it */
		if (FAILED(hr) || lngState != adStateClosed) conn->lpVtbl->Close(conn);
	}

	SAFE_RELEASE(rs);
	SAFE_RELEASE(conn);


	return;
}


/* ----------------------------------------------------------------------------------- */
#define TRACE printf

long ADOErrorInfo(ADOConnection * conn) {

	/* This function prints the error information after an ADO error.
	 * One function call may trigger multiple errors, each of which
	 * is listed. This returns the error number of the first error
	 * in the errors collection (the vast majority of the time there is 
	 * only one error) or -1 if this could not be obtained. */

	ADOErrors * pErrors = NULL;
	ADOError * pError = NULL;
	BSTR bStr = NULL;

	long i, nCount, lngErrNumber, resultErrNumber = -1;
	VARIANT errIndex;
	LPTSTR strTemp;
	HRESULT hr;

	errIndex.vt = VT_I4; /* Integer type variant */

	/* Get the errors collection */
	hr = conn->lpVtbl->get_Errors(conn,&pErrors);
	if (hr != S_OK) return resultErrNumber;

	/* Get the number of errors in the error collection */
	hr = pErrors->lpVtbl->get_Count(pErrors,&nCount);
	if (hr != S_OK) {
		SAFE_RELEASE(pErrors);
		return resultErrNumber;
	}

	/* Loop through the errors collection */
	for(i = 0; (i < nCount); i++ ) {

		TRACE( "\n *****Dumping ADO Error %d of %d*****", i+1, nCount );

		/* We fill a variant to indicate which error we want */
		errIndex.lVal = i;
		hr = pErrors->lpVtbl->get_Item(pErrors,errIndex, &pError );

		if (hr == S_OK) {
			/* We successfully retrieved an error object */

			hr = pError->lpVtbl->get_Source(pError,&bStr); /* Get info from error object */
			if (hr == S_OK) {
				if (BStr2TStr(bStr,&strTemp) == S_OK) { /* Convert info to TStr */
					TRACE( "\n Source       = %s", strTemp); /* Print info out */
					GlobalFree(strTemp); /* Free our TStr */
				}
				SFS(bStr); /* Free the BStr */
			}

			hr = pError->lpVtbl->get_Description(pError,&bStr);
			if (hr == S_OK) {
				if (BStr2TStr(bStr,&strTemp) == S_OK) {
					TRACE( "\n Description  = %s",  strTemp );
					GlobalFree(strTemp);
				}
				SFS(bStr); 
			}

			hr = pError->lpVtbl->get_HelpFile(pError,&bStr);
			if (hr == S_OK && bStr != NULL) {
				if (BStr2TStr(bStr,&strTemp) == S_OK) {
					TRACE( "\n Help File    = %s",  strTemp );
					GlobalFree(strTemp);
				}
				SFS(bStr);

				/* Only retrieve the help context if a help file was available */
				hr = pError->lpVtbl->get_HelpContext(pError,&lngErrNumber);
				if (hr == S_OK) {
					TRACE( "\n Help Context = %ld", lngErrNumber);
				}
 
			}

			hr = pError->lpVtbl->get_SQLState(pError,&bStr);
			if (hr == S_OK) {
				if (BStr2TStr(bStr,&strTemp) == S_OK) {
					TRACE( "\n SQL State    = %s",  strTemp );
					GlobalFree(strTemp);
				}
				SFS(bStr); 
			}


			hr = pError->lpVtbl->get_Number(pError,&lngErrNumber);
			if (hr == S_OK) {
				TRACE( "\n Number       = %ld", lngErrNumber);
				if (resultErrNumber == -1) resultErrNumber = lngErrNumber; /* We return the first error number */
			}


			hr = pError->lpVtbl->get_NativeError(pError,&lngErrNumber);
			if (hr == S_OK) {
				TRACE( "\n Native Error = %ld", lngErrNumber);
			}

			SAFE_RELEASE(pError);
		}
	}
	pErrors->lpVtbl->Clear(pErrors); /* Clear the errors collection */
	SAFE_RELEASE(pErrors);

	return resultErrNumber;
}


/* ----------------------------------------------------------------------------------- */
HRESULT GetFieldNames(ADORecordset * pRs, long  * plngFieldCount, LPTSTR ** strNameArray ) {

	/* This function retrieves the names of fields belonging to a recordset.
	 * It also retrieves the number of fields.
	 * The field names are placed in strNameArray and the number of fields is placed
	 * in plngFieldCount.
	 * The callee is required to free each name in the array and the array with
	 * GlobalFree after a successful call to this function. This can be performed
	 * using the CleanupStringArray() function.
	 * Returns E_FAIL or S_OK */

	ADOFields * pFields = NULL;
	ADOField * pField = NULL;
	BSTR bStr = NULL;

	long lngCount, successCnt = 0,idx;
	VARIANT vtItemIndex;
	HRESULT hr;

	vtItemIndex.vt = VT_I4; /* Integer type variant */

	/* Get the fields collection */
	hr = pRs->lpVtbl->get_Fields(pRs,&pFields);
	if (hr != S_OK) return E_FAIL;

	/* Get the number of fields in the fields collection */
	hr = pFields->lpVtbl->get_Count(pFields,&lngCount);
	if (hr != S_OK) {
		SAFE_RELEASE(pFields);
		return E_FAIL;
	}
	
	*plngFieldCount = lngCount;

	/* Alloc space for our names array */
	*strNameArray = GlobalAlloc(GMEM_FIXED, lngCount * sizeof(LPTSTR) );

	if (*strNameArray == NULL) {
		/* Memory failure */
		SAFE_RELEASE(pFields);
		return E_FAIL;
	}

	/* NULL out our array */
	ZeroMemory(*strNameArray, lngCount * sizeof(LPTSTR) );

	/* Loop through the fields collection */
	for (idx=0;idx < lngCount;idx++) {

		/* Retrieve a field object */
		vtItemIndex.lVal = idx;
		hr = pFields->lpVtbl->get_Item(pFields, vtItemIndex, &pField );

		if (hr == S_OK) {

			/* Get the name of the field, convert it to a TStr and place it in our array */
			hr = pField->lpVtbl->get_Name(pField,&bStr);
			if (hr == S_OK) {
				if (BStr2TStr(bStr,&((*strNameArray)[idx])) == S_OK) {
					successCnt++;			
				}
				SFS(bStr); 
			}

			SAFE_RELEASE(pField);
		}

	}

	SAFE_RELEASE(pFields);

	if (successCnt != lngCount) {

		/* Ah, shucks, we failed for one or more fields, let's cleanup */

		CleanupStringArray(*strNameArray,lngCount);
		*strNameArray = NULL;
		return E_FAIL;
	}

	return S_OK;
}


/* ----------------------------------------------------------------------------------- */
void CleanupStringArray(LPTSTR * strArray, long elementCnt) {

	/* Frees memory taken by each element in a string array
	 * and the array itself. */

	long idx;

	if (strArray != NULL) {

		for (idx=0;idx < elementCnt;idx++) {

			/* Free each string in the array */
			if ( strArray[idx] != NULL) GlobalFree( strArray[idx] );
		}

		/* Free the array */
		GlobalFree(strArray);
	}

}


/* ----------------------------------------------------------------------------------- */
/* Note: Uses the ADO_TRY and ADO_REPORT macros defined for the process function */
void Data2HTML() {

	/* This function is a virtual clone of the process() function. However, instead
	 * of outputting the data to the screen it outputs it to a file in HTML format.
	 * The file is created in the current directory with the file name 'webdata.htm'
	 * This file can then be viewed in a webbrowser. Like process, the data source
	 * can be changed by changing the connection string and the select statement. */

	ADOConnection * conn = NULL;
	ADORecordset * rs = NULL;
	BSTR bStr = NULL;
	LPTSTR * strArrayFieldNames = NULL;

	HRESULT hr;
	VARIANT_BOOL vblEOF;
	long lastADOErrorNum;
	long lngFieldCnt;
	LONG lngState;

	LPTSTR strTemp;
	VARIANT varFieldIndex, varFieldContents;
	long idx, recordNum = 1;
	FILE * hFile = NULL;

	/* Create our connection object */
	hr = CoCreateInstance(&CLSID_CADOConnection , NULL ,CLSCTX_INPROC_SERVER,
		&IID_IADOConnection ,(LPVOID*) &conn);

	if (hr != S_OK) return;

	/* Connect to the data source */
	bStr = SAS(L"Provider=Microsoft.Jet.OLEDB.4.0;Data Source=Whales.xls;Extended Properties=\"Excel 8.0;HDR=Yes;IMEX=1;\"" );
	ADO_TRY( conn->lpVtbl->Open(conn, bStr, bstrEmpty, bstrEmpty, 0) );
	SFS(bStr);

	/* Execute a select statement to retrieve a recordset */
	bStr = SAS(L"SELECT ID, Species, Population, Status FROM [Whales$] ORDER BY ID");
	ADO_TRY( conn->lpVtbl->Execute(conn, bStr, NULL,0,&rs) );
	SFS(bStr);

	/* Check that a valid, open recordset was returned */
	if ( rs == NULL  || rs->lpVtbl->get_State(rs,&lngState) != S_OK
			   || lngState == adStateClosed) {
		/* If a query that returns no data is executed a closed recordset is returned */
		goto cleanup;
	}

	/* Get field names and the number of fields */
	ADO_TRY( GetFieldNames(rs,&lngFieldCnt,&strArrayFieldNames) );

	varFieldIndex.vt = VT_I4; /* Integer type variant */


	/* Open our file for output */
	hFile = fopen("webdata.htm","w"); 
	if (hFile == NULL) goto cleanup;

	fprintf(hFile,"<HTML>\n<HEAD><TITLE>Data2HTML Generated Table</TITLE></HEAD>\n<BODY>\n"
			"<TABLE align=center CELLPADDING=10 BORDER=2 BORDERCOLOR=navy CELLSPACING=0>\n\n");
	
	/* Print out column headers */
	fprintf(hFile,"<TR>\n");
	for (idx=0;idx < lngFieldCnt;idx++) {
		fprintf(hFile,"<TH>%s</TH>\n", strArrayFieldNames[idx] );
	}
	fprintf(hFile,"</TR>\n\n");


	/* Loop through the records in the recordset */
	while (TRUE) {

		/* Check if we are at the end of the recordset */
		ADO_TRY( rs->lpVtbl->get_EOF(rs,&vblEOF) );
		if (vblEOF) break; /* No more records */

		fprintf(hFile,"<TR>\n");

		/* Loop through each field and output its contents */
		for (idx=0;idx < lngFieldCnt;idx++) {

			varFieldIndex.lVal = idx;

			/* get_Collect is a little known, undocumented shortcut to rs.fields.item(idx).value
			 * It allows us to skip getting the fields collection and then each field */
			ADO_REPORT( rs->lpVtbl->get_Collect(rs,varFieldIndex,&varFieldContents) );

			/* lastADOErrorNum is set in the ADO_REPORT macro */
			/* Convert the field contents into a TStr so we can print them */

			if (lastADOErrorNum == 0 && varFieldContents.vt == VT_NULL) strTemp = TEXT("#NULL#");
			else if (lastADOErrorNum == 0) strTemp = Variant2TStr(varFieldContents);
			else strTemp = NULL; /* Error */

			fprintf(hFile,"<TD>%s</TD>\n",(strTemp ? strTemp : TEXT("#ERROR#") ) );

			if ( strTemp != NULL && varFieldContents.vt != VT_NULL) GlobalFree(strTemp);
			VariantClear(&varFieldContents);
		}

		fprintf(hFile,"</TR>\n\n");

		/* Move to the next record */
		ADO_TRY( rs->lpVtbl->MoveNext(rs) );

		recordNum++;
	}


	fprintf(hFile,"</TABLE>\n</BODY>\n</HTML>");

cleanup:

	if (hFile != NULL) fclose(hFile);

	if (strArrayFieldNames != NULL) CleanupStringArray(strArrayFieldNames,lngFieldCnt);

	SFS(bStr);

	if (conn != NULL) {
		hr = conn->lpVtbl->get_State(conn,&lngState);
		/* Closing the connection also closes active recordsets associated with it */
		if (FAILED(hr) || lngState != adStateClosed) conn->lpVtbl->Close(conn);
	}

	SAFE_RELEASE(rs);
	SAFE_RELEASE(conn);


	return;

}


