/* ComHelper code by anonytmouse.
 * String conversion code is taken from code by Jeff Glatt which can be found
 * at the following URL:
 * http://codeguru.earthweb.com/ieprogram/cwebpage.html */

#include <windows.h>
#include <oleauto.h>
#include <objbase.h>

/* ----------------------------------------------------------------- */
HRESULT TStr2BStr(LPTSTR strIn, BSTR * strOut) {

	/* Converts a TStr into unicode if needed and allocates a BSTR
	 * It is the callee's repsonsibility to call SySFreeString(BSTR)
	 * when it is no longer required. Returns S_OK(0) or E_FAIL */

#ifndef UNICODE

	LPWSTR		buffer;
	DWORD		size;

	size = MultiByteToWideChar(CP_ACP, 0, strIn, -1, 0, 0);
	if ( !(buffer = (wchar_t *) GlobalAlloc(GMEM_FIXED, sizeof(wchar_t) * size) ) ) return E_FAIL;
	MultiByteToWideChar(CP_ACP, 0, strIn, -1, buffer, size);
	*strOut = SysAllocString(buffer);
	GlobalFree(buffer);

#else
	*strOut = SysAllocString(strIn);
#endif

	if (*strOut == NULL) return E_FAIL;

	return S_OK;
}


/* ----------------------------------------------------------------- */
HRESULT BStr2TStr(BSTR strIn, LPTSTR * strOut) {

	/* Converts a BStr to ansi if required and returns the result
	 * It is the callee's repsonsibility to free the resulting string
	 * with GlobalFree when it is no longer needed.
	 * If unicode is defined the function simply copies the contents
	 * of strIn to strOut. Returns S_OK(0) or E_FAIL */

	DWORD		size;

#ifndef UNICODE

	size = WideCharToMultiByte(CP_ACP, 0, strIn, -1, 0, 0,NULL,NULL);
	if ( !( *strOut = (char *) GlobalAlloc(GMEM_FIXED, sizeof(char) * size) ) ) return E_FAIL;
	WideCharToMultiByte(CP_ACP, 0, strIn, -1, *strOut, size,NULL,NULL);

#else

	size = lstrlen(strIn) + 1;
	if ( !( *strOut = (wchar_t *) GlobalAlloc(GMEM_FIXED, sizeof(wchar_t) * size) ) ) return E_FAIL;
	lstrcpyn(*strOut,strIn,size);

#endif

	return S_OK;
}


/* ----------------------------------------------------------------- */
LPTSTR Variant2TStr(VARIANT vtValue)
{
	/* Takes a variant and returns its value in a TSTR. It is the callee's
	 * responsibility to free the resulting TSTR with GlobalFree.
	 * Returns NULL if not successful. Note: Do not call GlobalFree
	 * on a null pointer. */

	VARIANT vtString;
	HRESULT hr;
	LPTSTR strTemp;

	VariantInit(&vtString);

	if(vtValue.vt == VT_BSTR) {
		hr = BStr2TStr(vtValue.bstrVal,&strTemp);
	}
	else {
		hr = VariantChangeType(&vtString, &vtValue, VARIANT_ALPHABOOL, VT_BSTR);

		if (hr != S_OK || vtString.vt != VT_BSTR) {
			VariantClear(&vtString);
			return NULL;
		}

		hr = BStr2TStr(vtString.bstrVal,&strTemp);
		VariantClear(&vtString);
	}

	if (hr == S_OK) return strTemp;
	else return NULL;

}


/* ----------------------------------------------------------------- */
BSTR bstrEmpty; /* This global variable can be used for optional BSTR arguments. */
VARIANT vtMissing; /* This global variable can be used for optional VARIANT arguments. */

#define SAFE_RELEASE(obj) if (obj != NULL) { \
				obj->lpVtbl->Release(obj); \
				obj = NULL; }

#define SAS(str) SysAllocString(str)
/* Next macro is surrounded with {} so 'if (cond) SFS(mybstr);' does not cause problems. */
#define SFS(bstr) { SysFreeString(bstr); bstr = NULL; }


/* ----------------------------------------------------------------- */
void InitComHelper() {
	/* Call this at the start of your program */

	CoInitialize(NULL);
	bstrEmpty = SysAllocString(L"");
	vtMissing.vt = VT_ERROR;
	vtMissing.scode = DISP_E_PARAMNOTFOUND;

}


/* ----------------------------------------------------------------- */
void DestroyComHelper() {
	/* Call this at the end of your program */

	SysFreeString(bstrEmpty);
	CoUninitialize();

}

